(defpackage bezier
  (:use :cl :vektor))

(in-package :bezier)

(defparameter *width* 256)
(defparameter *height* 256)

(declaim (optimize (debug 3)))

(defclass test-window (feuer:2d-window)
  ((quad :initform nil :accessor quad :initarg :quad))
  (:default-initargs :mode '(:double :rgb :depth)))


(defparameter data nil)


(defmethod glut:display ((window test-window))
  
  (gl:clear-color 0 0 0 1)
  (gl:clear :color-buffer-bit)
  (feuer:draw (quad window))
  (glut:swap-buffers))

(defmethod glut:idle ((window test-window))
  (glut:post-redisplay))


(defmethod glut:keyboard ((window test-window) key x y)
  (glut:destroy-current-window))


(defmethod glut:display-window :before ((window test-window))
  (setf (quad window)
        (make-instance 'feuer:2d-quad :texture
                       (feuer:make-texture *width* *height*
                                           :data (make-array (array-total-size data)
                                                             :displaced-to data))))
  (feuer:reinit (feuer:texture-of (quad window))))


(defparameter *points* (list (vec2 0 0)
                             (vec2 0 200)
                             (vec2 200 80)
                             (vec2 100 255)))

(defun casteljau (points lambda)
  (if (= 1 (length points))
      (first points)
      (casteljau (mapcar (lambda (a b)
                           (convex-combination a b lambda))
                         points
                         (rest points))
                 lambda)))

(defun casteljau-recursive (a b c d &optional (τ 0.5))
  (declare (optimize (debug 3)))
  (let* ((ab (convex-combination a b τ))
         (bc (convex-combination b c τ))
         (cd (convex-combination c d τ))
         (abc (convex-combination ab bc τ))
         (bcd (convex-combination bc cd τ))
         (abcd (convex-combination abc bcd τ)))
    (if (< (reduce #'+ (mapcar (lambda (p q)
                                 (distance p q))
                               (list a b c d)
                               (list b c d)))
           2)
        (list abcd)
        (append (casteljau-recursive a ab abc abcd τ)
                (list abcd)
                (casteljau-recursive abcd bcd cd d τ)))))

(defun casteljau-wrap (points)
  (declare (optimize (debug 3)))
  (destructuring-bind (a b c d) points
    (append (list a)
            (casteljau-recursive a b c d)
            (list d))))

(defun smoothstep (edge1 edge2 value)
  (declare (optimize (speed 3)))
  (let ((normalized (coerce  (min 1
                                  (max 0(/ (- value edge1)
                                           (- edge2 edge1))))
                             '(single-float 0.0))))
    (+ (* 6 (expt normalized 5))
       (* -15 (expt normalized 4))
       (* 10 (expt normalized 3)))))

(defun draw-bezier (&optional (radius 5))
  (declare (optimize (speed 3)))
  (setf data (make-array (list *width* *height* 4) :element-type 'unsigned-byte :initial-element 0))
  (let ((points (casteljau-wrap *points*)))
    (dotimes (x *width*)
      (format t "~a" x)
      (dotimes (y *height*)
        (let ((pos (vec2 x y)))
          (let ((dist (loop for p in points
                            minimize (distance-squared p pos))))
            (setf (aref data x y 1)
                  (round (* 255 (smoothstep (+ radius 1.0) (- radius 1.0) (sqrt dist)))))))))
    (loop for p in points
          do (setf (aref data
                         (round (x p))
                         (round (y p))
                         0)
                   255))))

(defun run ()
  (draw-bezier)
  (glut:display-window (make-instance 'test-window)))