(in-package :feuer)


(defun read-file (filename)
  (with-open-file (stream filename)
    (let ((strings ()))
      (do ((line (read-line stream nil)
                 (read-line stream nil)))
          ((null line))
        (setf strings
              (cons (concatenate 'string line (string #\Newline)) strings)))
      (reverse strings))))



(defclass* shader-program ()
  (vertex-shader-file 
   fragment-shader-file
   vertex-name
   fragment-name
   (uniforms (make-hash-table))
   (name 0)))


(defmethod reinit ((program shader-program))
  (let ((vertex-src (read-file (vertex-shader-file-of program)))
        (fragment-src (read-file (fragment-shader-file-of program))))
    (flet ((destroy-me ()
             (gl:delete-shader (vertex-name-of program))
             (gl:delete-shader (fragment-name-of program))
             (gl:delete-program (name-of program))))
      (when (and (name-of program)
                 (not (= 0 (name-of program))))
        (destroy-me))
      (setf (vertex-name-of program)
            (gl:create-shader :vertex-shader)
            (fragment-name-of program)
            (gl:create-shader :fragment-shader)
            (name-of program)
            (gl:create-program))
      (gl:shader-source (vertex-name-of program) vertex-src)
      (gl:shader-source (fragment-name-of program) fragment-src)
      (gl:compile-shader (vertex-name-of program))
      (gl:compile-shader (fragment-name-of program))
      (gl:attach-shader (name-of program) (vertex-name-of program))
      (gl:attach-shader (name-of program) (fragment-name-of program))
      (gl:link-program (name-of program))
      (let ((log (gl:get-program-info-log (name-of program))))
        (when (> (length log) 1)
          (debug-out log)
          (destroy-me)
          (setf (name-of program) 0)))
      (sb-ext:finalize program (lambda ()
                                 (destroy-me))))))

(defmethod use-program ((program shader-program))
  (when (and (name-of program)
             (not (= 0 (name-of program)))))
  (gl:use-program (name-of program)))

(defmethod set-uniform ((program shader-program) (uniform string) (value integer))
;  (print-vars uniform value)
  (gl:uniformi (gl:get-uniform-location (name-of program) uniform)
               value))

(defmethod set-uniform ((program shader-program) (uniform string) (value real))
  (gl:uniformf (gl:get-uniform-location (name-of program) uniform)
               value))

(defmethod set-uniform ((program shader-program) (uniform string) (value texture))
  (gl:uniformi (gl:get-uniform-location (name-of program) uniform)
               (name-of value)))

