(in-package :feuer)

;;
;; naming conventions:
;; * "paint" methods produce or modify arrays that
;;   can then be used by textures
;;

(defclass* texture ()
  ((name nil)
   (mipmap t)
   (internal-format :rgba)
   (format :rgba)
   (width)
   (height)
   (data)))

(defun format->size (format)
  ;todo: hack
  (length (string format)))

(defun initialize-texture (texture)

  ;bla
  (setf (data-of texture)
        (make-array (* (width-of texture) (height-of texture) (format->size (format-of texture)))
                    :element-type '(unsigned-byte 8))))


(defgeneric update-texture-data (texture array-or-surface))
(defmethod update-texture-data ((texture texture) (data vector))
  (assert (and (= 1 (length (array-dimensions data)))
               (= (* (width-of texture)
                     (height-of texture)
                     (format->size (format-of texture)))
                  (first (array-dimensions data)))))
  (setf (data-of texture) data))

(defmethod update-texture-data ((texture texture) (surface cairo:surface))
  (assert (and (= (height-of texture) (cairo:height surface))
               (= (width-of texture) (cairo:width surface))))
  (setf (data-of texture) (cairo:image-surface-get-data surface)))


(defun update-texture (texture &optional (data (data-of texture) data-supplied?))
  (declare (optimize (debug 3)))
  
  (when data-supplied? (update-texture-data texture data))
  (when (name-of texture)
    (let ((name (name-of texture))
          (width (width-of texture))
          (height (height-of texture)))
      (with-slots (mipmap internal-format format) texture
        (gl:bind-texture :texture-2d name)
        (gl:tex-image-2d :texture-2d 0 internal-format width height 0 format :unsigned-byte (data-of texture))
       ; (gl:tex-parameter :texture-2d :texture-wrap-t :clamp-to-edge)
       ; (gl:tex-parameter :texture-2d :texture-wrap-s :clamp-to-edge)
         (gl:tex-parameter :texture-2d :texture-wrap-t :mirrored-repeat)
         (gl:tex-parameter :texture-2d :texture-wrap-s :mirrored-repeat)
        (when mipmap
          (glu:build-2d-mipmaps :texture-2d internal-format width height format :unsigned-byte (data-of texture))
          
          (gl:tex-parameter :texture-2d :texture-min-filter :nearest)
          (gl:tex-parameter :texture-2d :texture-mag-filter :nearest )

    
      ;    (gl:tex-parameter :texture-2d :texture-mag-filter :nearest )

          )))))

(defmethod reinit ((texture texture))
  (when (name-of texture)
    (gl:delete-textures (list (name-of texture))))
  (let ((name (first (gl:gen-textures 1))))
    (setf (name-of texture) name)
    (sb-ext:finalize texture (lambda ()
                               (gl:delete-textures (list name)))))
 ; (format t "~a~%" ( name-of texture))
  (update-texture texture))

(defun make-texture (width height &key data (internal-format :rgba) (format :rgba) (create-mipmaps t))
  (make-instance 'texture
                 :data data
                 :width width
                 :height height
                 :mipmap create-mipmaps
                 :internal-format internal-format
                 :format format))


(defmethod bind-texture ((texture texture))
  (gl:bind-texture :texture-2d (name-of texture)))

(defun array-width (array)
  (first (array-dimensions array)))

(defun array-height (array)
  (second (array-dimensions array)))

(defparameter *gaussian-y* (make-array '(1 5)
                                       :element-type :single-float
                                       :initial-contents '(( 1 4 7 4 1))))
(defparameter *gaussian-x* (make-array '(5 1)
                                       :element-type :single-float
                                       :initial-contents '(( 1) ( 4) ( 7) ( 4) ( 1))))

#+nil(defun convolute (array kernel) ;only monochrome 2d arrays for now
  (assert (and (oddp (array-height kernel))
               (oddp (array-width kernel))))
  (let ((result (make-array (array-dimensions array) :element-type (array-element-type array)))
        (width (array-width array))
        (height (array-height array))
        (x-radius (/ (- (array-width kernel) 1) 2))
        (y-radius (/ (- (array-height kernel) 1) 2))
        (normalization-factor (/ (loop for x from 0 below (array-width kernel)
                                       summing (loop for y below (array-height kernel)
                                                     summing (aref kernel x y)))))
        (normalized-kernel (make-array (array-dimensions kernel) :element-type (array-element-type kernel))))
    (loop for x from 0 below (array-width kernel)
          do (loop for y below (array-height kernel)
                   do (setf (aref normalized-kernel x y)
                            (* (aref kernel x y) normalization-factor))))
    (loop for xa from 0 below width
          do (loop for ya from 0 below height
                   do (setf (aref result xa ya)
                            (loop for xk from (- x-radius) to x-radius
                                  summing (loop for yk from (- y-radius) to y-radius
                                                summing (let ((x (+ xa xk))
                                                              (y (+ ya yk)))
                                                          (if (or (< y 0)
                                                                  (< x 0)
                                                                  (>= x width)
                                                                  (>= y height))
                                                              0
                                                              (round (* (aref array x y)
                                                                        (aref normalized-kernel (+ x-radius xk) (+ y-radius yk)))))))))))
    result))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; font stuff

(defparameter *font* "Monospace")

(declaim (optimize (debug 3)))

(defun paint-text (text width height size
                   &key
                   (color cl-colors:+white+)
                   (bg-color cl-colors:+black+)
                   (bg-alpha (if (eq 'cl-colors:rgba (type-of bg-color))
                                 (cl-colors:alpha bg-color)
                                 1)))
  (let* ((surface (cairo:create-image-surface :argb32 width height))
         (cairo:*context* (cairo:create-context surface))
         (lines (split-sequence:split-sequence #\Newline text))
         (line-height (* 1.4 (cairo:text-height (cairo:get-text-extents "Igasjg|\"AKAJSHDdhkajsdhscheissendreck")))))
    (cairo:set-source-rgba (cl-colors:red bg-color)
                           (cl-colors:green bg-color)
                           (cl-colors:blue bg-color)
                           bg-alpha)
    (cairo:paint)
    (cairo:set-source-color color)
    (cairo:select-font-face *font* :normal :bold)
    (cairo:set-font-size size)

    (loop for line in lines
          for y from (* 3.5 line-height) by (* 1.1 line-height)
          do (cairo:move-to (* 0.1 line-height) y)
          do (cairo:show-text line))
    (make-array (* width height 4)
                :element-type '(unsigned-byte 8)
                :initial-contents (cairo:image-surface-get-data surface))))

(defun make-text-texture (text width height size &rest rest)
  (make-texture width height
                :data (apply #'paint-text text width height size rest)
                :format :bgra
                :internal-format :rgba
                :format :bgra))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; texture loading from png images
;;

(defun flat-array (array)
  (make-array (reduce #'* (array-dimensions array))
              :element-type (array-element-type array)
              :displaced-to array))

#+nil(defun make-png-texture (filename)
  (with-open-file (input filename
                         :element-type '(unsigned-byte 8))
    (let ((image (png:8-bit-image (png:decode input))))
      (destructuring-bind (width height colors) (array-dimensions image)
        ;(print-vars colors)
        (make-texture width height
                      :data (flat-array image)
                      :format (if (= colors 3)
                                  :rgb
                                  :rgba))))))