(in-package :feuer)


(eval-when (:compile-toplevel :load-toplevel :execute)
;;
;; two useful macro-writing macros from pcl:
;;

(defmacro with-gensyms ((&rest names) &body body)
  `(let ,(loop for n in names collect `(,n (gensym)))
     ,@body))

(defmacro once-only ((&rest names) &body body) ;stolen from pcl
  (let ((gensyms (loop for n in names collect (gensym))))
    `(let (,@(loop for g in gensyms collect `(,g (gensym))))
      `(let (,,@(loop for g in gensyms for n in names collect ``(,,g ,,n)))
        ,(let (,@(loop for n in names for g in gensyms collect `(,n ,g)))
           ,@body)))))



(defmacro with-matrix-mode ((mode) &body body)
  (with-gensyms (old-mode-name)
    `(let ((,old-mode-name (cffi:foreign-enum-keyword '%gl:enum
                                                      (gl:get-integer :matrix-mode))))
       (gl:matrix-mode ,mode)
       ,@body
       (gl:matrix-mode ,old-mode-name))))

(defmacro with-pushed-matrix ((mode) &body body)
  `(with-matrix-mode (,mode)
     (gl:with-pushed-matrix
         ,@body))))


;;
;; a mapcar that wraps
;;

(defun mapwrap (fn list &rest lists)
  (let* ((input-lists (cons list lists))
         (length (apply #'max (mapcar #'length input-lists)))
         (lists (make-list length :initial-element ())))
    (loop for i from 0 to (- length 1)
          do (setf lists
                   (mapcar (lambda (list input-list)
                             (or (cdr list)
                                 (copy-list input-list)))
                           lists
                           input-lists))
          collect (apply fn (mapcar #'first lists)))))



;;
;; cycle something by a given offset
;;

(defgeneric cycle  (thing &optional i j))
(defgeneric cycle! (thing &optional i j))

(defmethod cycle ((list cons) &optional (i 1) (j 0))
  (declare (ignore j))
  (let ((i (mod i (length list))))
    (append (subseq list i) (subseq list 0 i))))



(defun atan2-inner (x y) ;named -inner because interfaces might want to have the atan2 name
  "somewhat awful but stolen, tested and working atan2"
  (if (= y 0.0)
      (if (< x 0.0)
      (+ PI (/ PI 2.0))
    (/ PI 2.0))
    (if (= x 0.0)
    (if (> y 0.0)
        0.0
      PI)
      (let ((at (atan (/ x y))))
    (if (> x 0.0)
        (if (> y 0.0)
        at
          (+ PI at))
      (if (> y 0.0)
          (+ PI PI at)
        (+ PI at)))))))


(defun debug-out (string)
  (format t "~a~%" string)
  (force-output))

(defun choose (list)
  (when list
    (nth (random (length list)) list)))

(defun elt-mod (sequence index)
  (elt sequence (mod index (length sequence)))) 

(defun sqr (number)
  (* number number))


(defun transpose-list (list)
  (apply #'mapcar (cons 'list list)))

;(defun crop (number max &optional (min (- max)))
;  (max (min number max) min))

(defmacro print-vars (&rest rest)
  `(format t ,(format nil "~{~a:~~a~^, ~}~~%" rest) ,@rest)) ;self-explanatory :)

(defun mapvalue (fn hash-table)
  (maphash (lambda (k v)
             (declare (ignore k))
             (funcall fn v))
           hash-table))


;;;;;;;;;;;;;;;;;
;;
;; cl-opengl convenience
;;
;; 



(defun fill-list (list length fill-element)
  (append list (loop for i from (length list) below length
                     collect fill-element)))

(defmethod translate ((v vektor))
  (gl:translate (x v) (y v) (if (= (dimension v) 2) 0 (z v))))


(defmethod rotate ((angle number) (v vektor))
  (gl:rotate (* angle 180 (/ pi)) (x v) (y v) (z v)))

(defmethod rotate ((v vektor) (angle number))
  (gl:rotate (* angle 180 (/ pi)) (x v) (y v) (z v)))

(defmethod scale ((v vektor))
  (gl:scale (x v) (y v) (if (= (dimension v) 2) 1 (z v))))

(defmethod scale ((x number))
  (gl:scale x x x))



(defgeneric color (vector/list/color/r &optional alpha/g b a))

(defmethod color ((r number) &optional (g 1) (b 1) (a 1))
  (gl:color r g b a))


(defmethod color ((list cons) &optional _1 _2 _3)
  (apply #'gl:color list))

(defmethod color ((v vektor) &optional (a 1) _1 _2)
  (color (fill-list (data v) 4 1)))

(defmethod color ((color cl-colors:rgb) &optional (a 1) _1 _2)
  (gl:color (cl-colors:red color)
            (cl-colors:green color)
            (cl-colors:blue color)
            a))


(defmethod color ((color cl-colors:rgba) &optional _1 _2 _3)
  (gl:color (cl-colors:red color)
            (cl-colors:green color)
            (cl-colors:blue color)
            (cl-colors:alpha color)))
