(in-package :feuer)

;todo: mehrere sorten windows machen fuer overlay, arcball, ...

(defparameter *supported-chars* "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()_+,.;:<>/?\\|[]{}~`'\" ")

(defclass* overlay-window (glut:window)
  ((overlay-elements ())))


(defmethod initialize-instance :after ((window overlay-window) &key (overlay-elements nil))
  (when overlay-elements
    (dolist (elem overlay-elements)
      (setf (window-of elem) window))
    (setf (overlay-elements-of window) overlay-elements)))

(defclass* overlay-element ()
  ((pos (make-point 0 0 0))   ;todo: sort by z coordinate if given
   window
   (pos-style :fixed)
   (pos-anchor :top-left) ;one of {top,bottom}-{left,right},center
   (active? t)
   (bg-alpha 0.5)
   size
   ;(size-style :fixed)
   quad))

(defun make-single-quad-mesh (texture-width texture-height &key (texture-data nil) (texture-format :bgra))
  (let ((mesh (make-instance 'mesh
                               :patches
                               (list (make-instance 'quad-patch)))))
    (setf (polygons-of (first (patches-of mesh)))
          (list (make-instance 'quad :vertices
                                     (apply #'mapcar (lambda (x y z u v)
                                                       (let ((vertex (make-vertex x y z)))
                                                         (setf (tex-coord-of vertex
                                                                             (first (patches-of mesh)))
                                                               (make-point u v))
                                                         vertex))
                                            (transpose-list'((-0.5 -0.5  0    0  0)
                                                             (-0.5  0.5  0    0  1)
                                                             ( 0.5  0.5  0    1  1)
                                                             ( 0.5 -0.5  0    1  0))))))
          (texture-of (first (patches-of mesh)))
                (make-texture texture-width texture-height
                              :data (or texture-data
                                        (make-array
                                         (* texture-width texture-height 4) :element-type '(unsigned-byte 8) :initial-element 0))
                              :format texture-format))
    (update-topology mesh)
    mesh))

(defmethod initialize-instance :after ((overlay-element overlay-element) &key (width nil width?) (height nil height?))
  (when (and width? height?)
    (setf (size-of overlay-element) (make-point width height)))
  (setf (quad-of overlay-element)
        (make-single-quad-mesh (round (x (size-of overlay-element)))
                               (round (y (size-of overlay-element))))))

(defmethod draw ((overlay-element overlay-element) &key)
  (when (active?-of overlay-element)
    (with-pushed-matrix (:modelview)
      (apply #'gl:translate (case (pos-anchor-of overlay-element)
                              (:middle       '( 0    0    0))
                              (:top-left     '(-0.5 -0.5  0))
                              (:top-right    '( 0.5 -0.5  0))
                              (:bottom-left  '(-0.5  0.5  0))
                              (:bottom-right '( 0.5  0.5  0))))
      (gl:scale (/  (glut:width (window-of overlay-element)))
                (/  (glut:height (window-of overlay-element)))
                1)
      (translate (pos-of overlay-element))
      (scale (size-of overlay-element))
      (draw (quad-of overlay-element)))))



(defclass* text-overlay (overlay-element)
  ((text "uninitalized..." :reader text)
   (text-size 16)
   (font "Helvetica")
   (color cl-colors:+snow+)))

(defmethod draw :before ((text text-overlay) &key)
  (color (color-of text)))

(defmethod (setf text-of) (new-text (overlay text-overlay))
  (setf (slot-value overlay 'text) new-text)
  (let ((texture (texture-of (first (patches-of (quad-of overlay))))))
    (update-texture texture
                    (paint-text new-text (width-of texture) (height-of texture) (text-size-of overlay)
                                     :bg-alpha (bg-alpha-of overlay)))))

(defmethod initialize-instance :after ((text-overlay text-overlay) &key (text nil text-supplied?))
  (when text-supplied?
    (setf (text-of text-overlay) text)))

(defmethod reinit ((overlay text-overlay))
  (reinit (texture-of (first (patches-of (quad-of overlay))))))

(defmethod glut:display-window :before ((window overlay-window))
  ;(debug-out "reiniting overlays")
  (mapcan (lambda (overlay)
            (setf (window-of overlay) window))
          (overlay-elements-of window))
  (mapcan #'reinit (overlay-elements-of window)))


(defmethod glut:display ((window glut:window))
  (glut:swap-buffers))

(defmethod glut:display ((window overlay-window))
  (gl:enable :blend)
  (gl:blend-func :src-alpha
                 :one-minus-src-alpha)
  (gl:disable :depth-test)
  (with-pushed-matrix (:projection)
    (gl:load-identity)
    (gl:ortho 0.5 -0.5 0.5 -0.5 -100 100)
    (with-pushed-matrix (:modelview)
      (gl:load-identity)
      (glu:look-at 0 0 -1  0 0 0  0 1 0)
      (mapcan #'draw (overlay-elements-of window))))
  (call-next-method))


;;
;; skyboxes
;;
;; sehr unfertig....

(defclass* skybox-window (glut:window)
  ((skybox-filename "test.png")
   skybox-texture))

(defmethod initialize-instance :after ((window skybox-window) &key (skybox-filename "skybox.png"))
  (setf (skybox-texture-of window) (feuer::make-png-texture skybox-filename)))

(defmethod glut:display ((window skybox-window))
  ;(call-next-method) ;eigentlich sollte  glut:display method combination progn haben.
  ;(debug-out "foo!")
  (glut:solid-teapot 1))
